# Copyright (c) 2024 - 2025 Noppanut Ploywong (@noppanut15) <noppanut.connect@gmail.com>
# Apache License 2.0 (see LICENSE file or http://www.apache.org/licenses/LICENSE-2.0)


"""
Generic video creator module to create a video from an array input.
"""

import os.path
from typing import Tuple, cast, Union
from moviepy import TextClip, VideoClip, concatenate_videoclips
from tqdm import tqdm
from depthviz.video.logger import DepthVizProgessBarLogger

# Default values
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_FONT = os.path.abspath(
    os.path.join(BASE_DIR, "../assets/fonts/Open_Sans/static/OpenSans-Bold.ttf")
)
DEFAULT_VIDEO_SIZE = (960, 540)
DEFAULT_VIDEO_SIZE_FOR_TESTING = (640, 360)
DEFAULT_BG_COLOR = "black"
DEFAULT_STROKE_WIDTH = 5


class OverlayVideoCreatorError(Exception):
    """Base class for exceptions in this module."""


class VideoNotRenderError(OverlayVideoCreatorError):
    """Exception raised for video not rendered errors."""


class VideoFormatError(OverlayVideoCreatorError):
    """Exception raised for invalid video format errors."""


class OverlayVideoCreator:
    """
    Generic class to create an overlay video from an array input.
    """

    def __init__(
        self,
        font: str = DEFAULT_FONT,
        interline: int = -20,
        color: str = "white",
        bg_color: str = DEFAULT_BG_COLOR,
        stroke_color: str = "black",
        stroke_width: int = DEFAULT_STROKE_WIDTH,
        align: str = "center",
        size: Tuple[int, int] = DEFAULT_VIDEO_SIZE_FOR_TESTING,
        bitrate: str = "5000k",
        fps: int = 25,
    ):
        self.font = font
        self.fontsize = int(
            size[1] * 120 / 360
        )  # 120 is the default font size for 640x360 resolution
        self.interline = interline
        self.color = color
        self.bg_color = bg_color
        self.stroke_color = stroke_color
        self.stroke_width = stroke_width
        self.align = align
        self.size = size
        self.bitrate = bitrate
        self.fps = fps
        self.progress_bar_logger_config = {
            "unit": "f",
            "color": "#3982d8",
            "ncols": 70,
        }

        # Validate the font file
        self.__font_validate()
        # Validate the background color
        self.__bg_color_validate()
        # Validate the stroke width
        if not isinstance(stroke_width, int) or stroke_width < 0:
            raise OverlayVideoCreatorError(
                "Invalid stroke width; must be a positive number."
            )

    def render_text_video(
        self,
        text_list: list[dict[str, Union[str, float]]],
        progress_bar_desc: str = "Rendering",
    ) -> VideoClip:
        """
        Creates a video from an array input.

        Args:
            text_list: A list of dictionaries containing the text and duration for each frame.
            progress_bar_desc: The description for the progress bar.

        Returns:
            The processed video.
        """
        # Create a text clip for each text value and track the progress with a progress bar
        clips = []
        clip_count = len(text_list)
        for i in tqdm(
            iterable=range(clip_count),
            desc=progress_bar_desc,
            colour=str(self.progress_bar_logger_config["color"]),
            unit=str(self.progress_bar_logger_config["unit"]),
            ncols=cast(int, self.progress_bar_logger_config["ncols"]),
            bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} ({remaining} remaining)",
            leave=False,
        ):

            clip = TextClip(
                text=text_list[i]["text"],
                font=self.font,
                font_size=self.fontsize,
                interline=self.interline,
                color=self.color,
                bg_color=self.bg_color,
                stroke_color=self.stroke_color,
                stroke_width=self.stroke_width,
                text_align=self.align,
                size=self.size,
                duration=text_list[i]["duration"],
            )
            clips.append(clip)

        # Concatenate all the clips into a single video
        full_video = concatenate_videoclips(clips)
        return full_video

    def save(
        self, video: VideoClip, path: str, progress_bar_desc: str = "Exporting"
    ) -> None:
        """
        Saves the video to a file.

        Args:
            video: The video to save.
            path: The path to save the video (expected file format: .mp4).
        """
        parent_dir = os.path.dirname(path)
        if parent_dir == "":
            parent_dir = "./"
        if os.path.exists(parent_dir):
            if os.path.isdir(path):
                raise NameError(
                    f"{path} is a directory, please add a file name to the path. \
                        (e.g., path/to/mydepth_video.mp4)"
                )
            if video is not None:
                if not path.endswith(".mp4"):
                    raise VideoFormatError(
                        "Invalid file format: The file format must be .mp4"
                    )
                video.write_videofile(
                    path,
                    fps=self.fps,
                    bitrate=self.bitrate,
                    logger=DepthVizProgessBarLogger(
                        description=progress_bar_desc,
                        unit=cast(str, self.progress_bar_logger_config["unit"]),
                        color=cast(str, self.progress_bar_logger_config["color"]),
                        ncols=cast(int, self.progress_bar_logger_config["ncols"]),
                    ),
                )
            else:
                raise VideoNotRenderError(
                    "Cannot save video because it has not been rendered yet."
                )
        else:
            raise FileNotFoundError(f"Parent directory does not exist: {parent_dir}")

    def __font_validate(self) -> None:
        """
        Validates the font file.
        """
        # Check if the font file exists
        if not os.path.exists(self.font):
            raise OverlayVideoCreatorError(f"Font file not found: {self.font}")

        # Check if the font file is a file
        if not os.path.isfile(self.font):
            raise OverlayVideoCreatorError(
                f"Font you provided is not a file: {self.font}"
            )

        # Check if the font file is a valid font file
        try:
            TextClip(font=self.font, text="Test", font_size=1)
        except ValueError as e:
            raise OverlayVideoCreatorError(
                f"Error loading font file: {self.font}, "
                "make sure it's a valid font file (TrueType or OpenType font)."
            ) from e

    def __bg_color_validate(self) -> None:
        """
        Validates the background color.
        """
        # Check if the background color is a valid color
        try:
            TextClip(
                text="Test",
                font=self.font,
                font_size=self.fontsize,
                bg_color=self.bg_color,
            )
        except ValueError as e:
            raise OverlayVideoCreatorError(
                f"Invalid background color: {self.bg_color}"
            ) from e



"""
This module contains the DiveLogParser base class 
which is used to parse a dive log file containing depth data.
"""

from abc import ABC, abstractmethod

# Constants
ASSUMED_DESCENT_RATE = 1  # meters per second
ASSUMED_ASCENT_RATE = 1  # meters per second
ASCENT_CUTOFF_DEPTH = 1  # meters


class DiveLogParserError(Exception):
    """Base class for exceptions in this module."""


class DiveLogFileNotFoundError(DiveLogParserError):
    """Exception raised for file not found errors."""


class InvalidTimeValueError(DiveLogParserError):
    """Exception raised for invalid time value errors."""


class InvalidDepthValueError(DiveLogParserError):
    """Exception raised for invalid depth value errors."""


class EmptyFileError(DiveLogParserError):
    """Exception raised for empty file errors."""


class DiveLogParser(ABC):
    """
    A class to parse a dive log file containing depth data.
    """

    def __init__(self, depth_mode: str) -> None:
        """
        Initializes the DiveLogParser object.
        """

        self.time_data: list[float] = []
        self.depth_data: list[float] = []
        self.depth_mode = depth_mode

    @abstractmethod
    def parse(self, file_path: str) -> None:
        """
        Parses a dive log file containing depth data.

        Parameters:
        file_path (str): The path to the dive log file to be parsed.
        """

    @abstractmethod
    def get_time_data(self) -> list[float]:
        """
        Returns the time data parsed from the dive log file.

        Returns:
        list[float]: The time data parsed from the dive log file.
        """

    @abstractmethod
    def get_depth_data(self) -> list[float]:
        """
        Returns the depth data parsed from the dive log file.

        Returns:
        list[float]: The depth data parsed from the dive log file.
        """

    def ensure_zero_depth_at_surface(self) -> None:
        """
        Ensures that the depth data starts and ends at zero.
        """
        # Step 1: Ensure that the depth data STARTS at zero
        first_depth = self.depth_data[0]
        first_time = self.time_data[0]
        if first_depth > 0:
            self.depth_data.insert(0, 0)
            duration = (1 / ASSUMED_DESCENT_RATE) * first_depth
            self.time_data.insert(0, first_time - duration)

        # Step 2: Ensure that the depth data END at zero
        last_depth = self.depth_data[-1]
        last_time = self.time_data[-1]
        if last_depth > 0:
            # Clean up the depth data to remove any values close to zero
            # (between 0 and ASCENT_CUTOFF_DEPTH)
            for idx in range(len(self.depth_data) - 1, -1, -1):
                if self.depth_data[idx] < ASCENT_CUTOFF_DEPTH:
                    self.depth_data.pop(idx)
                    self.time_data.pop(idx)
                else:
                    break
            # Add a zero depth value at the end
            last_time = self.time_data[-1]
            last_depth = self.depth_data[-1]
            self.depth_data.append(0)
            duration = (1 / ASSUMED_ASCENT_RATE) * last_depth
            self.time_data.append(last_time + duration)

        # Step 3: Ensure the time data is positive (by starting at zero)
        self.time_data = [time - self.time_data[0] for time in self.time_data]

    def depth_mode_execute(self) -> None:
        """
        Executes the depth mode operation on the depth and time data.
        """
        if self.depth_mode == "zero-based":
            self.ensure_zero_depth_at_surface()
        elif self.depth_mode == "raw":
            # No operation needed
            pass
        else:
            raise DiveLogParserError("Invalid depth mode")



# Copyright (c) 2024 - 2025 Noppanut Ploywong (@noppanut15) <noppanut.connect@gmail.com>
# Apache License 2.0 (see LICENSE file or http://www.apache.org/licenses/LICENSE-2.0)


"""
Module to create a video that reports the depth in meters from an array input.
"""

from typing import Union
from moviepy import VideoClip
from depthviz.video.video_creator import OverlayVideoCreator, OverlayVideoCreatorError
from depthviz.optimizer.linear_interpolation import (
    LinearInterpolationDepth,
    LinearInterpolationDepthError,
)


class DepthReportVideoCreatorError(OverlayVideoCreatorError):
    """Base class for exceptions in this module."""


class DepthReportVideoCreator(OverlayVideoCreator):
    """
    Class to create a video that reports the depth in meters from an array input.
    """

    def __clip_duration_in_seconds(
        self, current_pos: int, time_data: list[float]
    ) -> float:
        """
        Returns the total duration of the video in seconds.

        Args:
            current_pos: The current position in the array.
            time_data: An array of time values in seconds.

        Returns:
            The total duration of the video in seconds.
        """
        if current_pos == len(time_data) - 1:
            # If it's the last element, return the difference between the last two elements
            return abs(time_data[current_pos] - time_data[current_pos - 1])
        # Otherwise, return the difference between the current and next element
        return abs(time_data[current_pos + 1] - time_data[current_pos])

    def render_depth_report_video(
        self,
        time_data: list[float],
        depth_data: list[float],
        decimal_places: int = 0,
        minus_sign: bool = True,
    ) -> VideoClip:
        """
        Creates a video that reports the depth in meters from an array input.

        Args:
            time_data: An array of time values in seconds.
            depth_data: An array of depth values in meters.
            decimal_places: The number of decimal places to round the depth values to.
            minus_sign: A boolean value to determine if the minus sign should be displayed.

        Returns:
            The processed video.
        """
        # Check the decimal places value
        if (
            not isinstance(decimal_places, int)
            or decimal_places < 0
            or decimal_places > 2
        ):
            raise DepthReportVideoCreatorError(
                "Invalid decimal places value; must be a number between 0 and 2."
            )
        # Interpolate the depth data
        try:
            interpolated_depth = LinearInterpolationDepth(
                times=time_data, depths=depth_data, fps=self.fps
            )
            interpolated_depths = interpolated_depth.get_interpolated_depths()
            interpolated_times = interpolated_depth.get_interpolated_times()

            depth_frame_list: list[dict[str, Union[str, float]]] = []
            clip_count = len(interpolated_times)
            for i in range(clip_count):
                duration = self.__clip_duration_in_seconds(i, interpolated_times)
                if decimal_places == 0:
                    rounded_current_depth = round(interpolated_depths[i])
                    if rounded_current_depth == 0:
                        text = "0m"
                    else:
                        text = f"{'-' if minus_sign else ''}{rounded_current_depth}m"
                else:
                    current_depth = round(interpolated_depths[i], decimal_places)
                    if current_depth == 0:
                        text = f"{0:.{decimal_places}f}m"
                    else:
                        text = f"{'-' if minus_sign else ''}{current_depth:.{decimal_places}f}m"
                depth_frame_list.append({"text": text, "duration": duration})

            full_video = super().render_text_video(depth_frame_list)
            return full_video
        except LinearInterpolationDepthError as e:
            raise DepthReportVideoCreatorError(f"Interpolation Error; ({e})") from e



# Copyright (c) 2024 - 2025 Noppanut Ploywong (@noppanut15) <noppanut.connect@gmail.com>
# Apache License 2.0 (see LICENSE file or http://www.apache.org/licenses/LICENSE-2.0)


"""
Version information for depthviz, will be replaced during the release process.
"""

# These version placeholders will be replaced later during substitution.
__version__ = "0.5.3b2"
__version_tuple__ = (0, 5, "3b2")



# Copyright (c) 2024 - 2025 Noppanut Ploywong (@noppanut15) <noppanut.connect@gmail.com>
# Apache License 2.0 (see LICENSE file or http://www.apache.org/licenses/LICENSE-2.0)


"""
This module is the main entry point for the command line interface (CLI) of the DepthViz package.
"""

import sys
from depthviz.main import run

if __name__ == "__main__":
    sys.exit(run())



# Copyright (c) 2024 - 2025 Noppanut Ploywong (@noppanut15) <noppanut.connect@gmail.com>
# Apache License 2.0 (see LICENSE file or http://www.apache.org/licenses/LICENSE-2.0)


"""
Module to create a timer video.
"""

import math
from typing import Union
from moviepy import VideoClip
from depthviz.video.video_creator import (
    OverlayVideoCreator,
    OverlayVideoCreatorError,
    DEFAULT_FONT,
    DEFAULT_BG_COLOR,
    DEFAULT_STROKE_WIDTH,
    DEFAULT_VIDEO_SIZE,
)


class TimeReportVideoCreatorError(OverlayVideoCreatorError):
    """Base class for exceptions in this module."""


class TimeReportVideoCreator(OverlayVideoCreator):
    """
    Class to create a video that reports the depth in meters from an array input.
    """

    def __init__(
        self,
        font: str = DEFAULT_FONT,
        bg_color: str = DEFAULT_BG_COLOR,
        stroke_width: int = DEFAULT_STROKE_WIDTH,
        size: tuple[int, int] = DEFAULT_VIDEO_SIZE,
    ):
        super().__init__(
            font=font,
            bg_color=bg_color,
            stroke_width=stroke_width,
            size=size,
            fps=4,  # Set the frame rate to 4 fps
        )

    def _convert_time_to_text(self, time: float) -> str:
        """
        Converts a time value to a text string.

        Args:
            time: The time value in seconds.

        Returns:
            The time value as mm:ss.
        """
        minutes = int(time // 60)
        seconds = int(time % 60)
        return f"{minutes:02d}:{seconds:02d}"

    def render_time_report_video(
        self,
        time_data: list[float],
    ) -> VideoClip:
        """
        Creates a video that reports the depth in meters from an array input.

        Args:
            time_data: An array of time values in seconds.

        Returns:
            The processed video.
        """
        # Check the time data is not empty
        if not time_data:
            raise TimeReportVideoCreatorError("The time data is empty.")

        # Check the time data is positive
        if any(time < 0 for time in time_data):
            raise TimeReportVideoCreatorError("The time data contains negative values.")

        # Create a list of time frames
        time_frame_list: list[dict[str, Union[str, float]]] = []
        start = 0
        end = math.ceil(time_data[-1] - time_data[0])
        for time in range(start, end + 1):
            for _ in range(self.fps):
                time_frame_list.append(
                    {"text": self._convert_time_to_text(time), "duration": 1 / self.fps}
                )
        full_video = super().render_text_video(time_frame_list)
        return full_video

    def to_time_output_path(self, path: str) -> str:
        """
        Convert a video path to a time output path by adding a suffix `_time`.

        Args:
            path: The video path.

        Returns:
            The time output path.
        """
        path_split = path.split(".")
        path_split[-2] += "_time"
        return ".".join(path_split)

    def save(
        self, video: VideoClip, path: str, progress_bar_desc: str = "Exporting"
    ) -> None:
        """
        Save the video to a file.

        Args:
            video: The video to save.
            path: The path to save the video to.
        """
        # Add a suffix `_time` to the file name
        path = self.to_time_output_path(path)
        super().save(video=video, path=path, progress_bar_desc=progress_bar_desc)
